unit IC_LMM;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	IC Location Manager Module
	--------------------------
	This file implements a Location Manager Module (LMM) for switching between
	IC profiles.  The implementation follows fairly standard LMM lines.
*)

interface

	uses
		Components;

{$main}
	function Main(var params : ComponentParameters; storage : Ptr) : ComponentResult;

implementation

	uses
		Resources,
		LocationManager,
		Script,
		TextUtils,

		PascalA4,

		InternetConfig,
		
		ICDebug,
		ICCommonSubs,
		ICStringLists;

	(* ***** Broken Apple Interfaces ***** *)

	// Location Manager selectors that should be defined in
	// "LocationManager.p" but aren't because, in the C interfaces, they
	// are defined in "LocationManager.k.h" (WTF does the ".k" mean), which
	// is not ported to Pascal, so the constants never made it to Pascal
	// either.  I've already filed this as a bug.
	
	const
		kALMGetCurrentSelect = $0000;
		kALMSetCurrentSelect = $0001;
		kALMCompareSettingSelect = $0002;
		kALMEditSettingSelect = $0003;
		kALMDescribeSettingsSelect = $0004;
		kALMDescribeErrorSelect = $0005;
		kALMImportExportSelect = $0006;
		kALMGetScriptInfoSelect = $0007;
		kALMGetInfoSelect = $0008;
		
		kALMFirstSelect = kALMGetCurrentSelect;
		kALMLastSelect = kALMGetInfoSelect;

	(* ***** Resources ***** *)
	
	const
		// 'TEXT'
		
		rInfoText = 128;
		
		// 'STR#'
		
		rMiscStrings = 128;
			strInternetConfigSet = 1;
			strNoICErr  = 2;
			strInsufficientICErr = 3;
			strNoMemoryErr = 4;
			strCantFindICErr = 5;
			strCantHackIt = 6;
			strProfileNotFoundErr = 7;
			strInvalidSetting = 8;
			strSettingTooOld = 9;
			strSettingTooNew = 10;
			
	(* ***** Misc Constants ***** *)

	const
		kICLMMCreator = 'IClM';
	
	const
		memFullFakeErr = 1;
		afpItemNotFoundFakeErr = 2;
		noPortFakeErr = 3;
		invalidSettingFakeErr = 4;
		settingTooOldFakeErr = 5;
		settingTooNewFakeErr = 6;
			
	(* ***** Globals ***** *)

	type
		ComponentGlobals = 
			record
				self : ComponentInstance;
				target : ComponentInstance;
				inst : ICInstance;
				ourResFile : integer;
			end;
		ComponentGlobalsPtr = ^ComponentGlobals;
	
	(* ***** Component Manager Routines ***** *)
	
	(* Routines that are generally implemented by components. *)

	function ICLMMCanDo(globals : ComponentGlobalsPtr; selector : integer) : ComponentResult;
	begin
		{$unused globals}
		case selector of
			kComponentVersionSelect,
				kComponentCanDoSelect,
					kComponentOpenSelect,
					kComponentCloseSelect,
					kComponentTargetSelect,
					kALMFirstSelect..kALMLastSelect:
				ICLMMCanDo := 1;
			otherwise
				ICLMMCanDo := 0;
		end; (* case *)
	end; (* ICLMMCanDo *)

	function ICLMMClose(globals : ComponentGlobalsPtr; self : ComponentInstance) : ComponentResult;
		var
			err : ComponentResult;
			junk : OSStatus;
	begin
		{$unused self}
		err := noErr;
		if globals <> nil then begin
			if globals^.inst <> nil then begin
				junk := ICStop(globals^.inst);
				globals^.inst := nil;
				ICAssert(junk = noErr);
			end; (* if *)
			if globals^.ourResFile <> 0 then begin
				junk := CloseComponentResFile(globals^.ourResFile);
				globals^.ourResFile := 0;
				ICAssert(junk = noErr);
			end; (* if *)
			DisposePtr(Ptr(globals));
			ICAssert(MemError() = noErr);
		end; (* if *)
		ICLMMClose := err;
	end; (* ICLMMClose *)
	
	function ICLMMOpen(globals : ComponentGlobalsPtr; self : ComponentInstance) : ComponentResult;
		var
			err: ComponentResult;
			junk : OSStatus;
	begin
		// Create our globals.
		
		globals := ComponentGlobalsPtr(NewPtrClear(sizeof(ComponentGlobals)));
		err := CheckMemError(globals);

		// Initialise our globals and tell the Component Manager about them.
		
		if err = noErr then begin
			globals^.self := self;
			globals^.target := self;
			globals^.inst := nil;
			globals^.ourResFile := 0;
			SetComponentInstanceStorage(self, Handle(globals));
		end; (* if *)

		// Clean up on failure.
				
		if err <> noErr then begin
			junk := ICLMMClose(globals, self);
			ICAssert(junk = noErr);
		end; (* if *)

		ICLMMOpen := noErr;
	end; (* ICLMMOpen *)

	function ICLMMTarget (globals: ComponentGlobalsPtr; newTarget: ComponentInstance): ComponentResult;
	begin
		globals^.target := newTarget;
		ICLMMTarget := noErr;
	end; (* ICLMMTarget *)

	(* ***** Setting Data Type ***** *)
	
	type
{$PUSH}
{$ALIGN MAC68K}
		ICSetting =
			record
				magic : OSType;					// must be kICLMMCreator
				version : longint;
				profileID : ICProfileID;
				profileNameScript : ScriptCode;
				profileName : Str255;
			end;
		ICSettingPtr = ^ICSetting;
		ICSettingHandle = ^ICSettingPtr;
{$ALIGN RESET}
{$POP}
	const
		kMinimumSettingSize = 8;				// sizeof(magic) + sizeof(version)
		kCurrentSettingVersion = 0;

	(* ***** Utility Routines ***** *)
	
	function StartUsingIC(globals : ComponentGlobalsPtr) : OSStatus;
		var
			err : OSStatus;
	begin
		err := noErr;
		if globals^.inst = nil then begin
			err := ICStart(globals^.inst, kICLMMCreator);
			if err = noErr then begin
				err := ICFindConfigFile(globals^.inst, 0, nil);
			end; (* if *)
		end; (* if *)
		StartUsingIC := err;
	end; (* StartUsingIC *)
	
	function BeginUsingResources(globals : ComponentGlobalsPtr; var oldResFile : integer) : OSStatus;
		var
			err : OSStatus;
	begin
		err := noErr;
		oldResFile := CurResFile;
		if globals^.ourResFile = 0 then begin
			globals^.ourResFile := OpenComponentResFile(Component(globals^.self));
			err := ResError;
		end else begin
			UseResFile(globals^.ourResFile);
			ICAssert(ResError = noErr);
		end; (* if *)
		BeginUsingResources := err;
	end; (* BeginUsingResources *)
	
	procedure EndUsingResources(oldResFile : integer);
	begin
		if oldResFile <> 0 then begin
			UseResFile(oldResFile);
			ICAssert(ResError = noErr);
		end; (* if *)
	end; (* EndUsingResources *)
	
	function ValidateSetting(setting : ICSettingHandle) : OSStatus;
		var
			err : OSStatus;
	begin
		err := noErr;
		if (setting = nil) | (setting^ = nil) then begin
			err := paramErr;
		end; (* if *)
		if err = noErr then begin
			if (GetHandleSize(Handle(setting)) < kMinimumSettingSize) | (setting^^.magic <> kICLMMCreator) then begin
				err := invalidSettingFakeErr;
			end; (* if *)
		end; (* if *)
		if err = noErr then begin

			// If you want anything else than this, you must program
			// special logic to convert settings from one version to another.
			// Of course, if you do this you must make sure that all the call
			// sites are happy with you munging the setting.
			
			if setting^^.version < kCurrentSettingVersion then begin
				err := settingTooOldFakeErr;
			end; (* if *)
			if setting^^.version > kCurrentSettingVersion then begin
				err := settingTooNewFakeErr;
			end; (* if *)
		end; (* if *)
		if err = noErr then begin
			if (GetHandleSize(Handle(setting)) <> sizeof(ICSetting)) | (setting^^.profileID = kICNilProfileID) then begin
				err := invalidSettingFakeErr;
			end; (* if *)
		end; (* if *)
		ValidateSetting := err;
	end; (* ValidateSetting *)
	
	function ProfileNamesEqual(const name1 : Str255; const name2 : Str255) : Boolean;
		var
			lang : LangCode;
	begin
		lang := GetScriptVariable(smSystemScript, smScriptLang);
		ProfileNamesEqual := StringOrder(name1, name2, smSystemScript, smSystemScript, lang, lang) = 0;
	end; (* ProfileNamesEqual *)

	function ResolveSettingByName(globals : ComponentGlobalsPtr;
							profileName : Str255;
							var trueProfileID : ICProfileID) : OSStatus;
		var
			err : OSStatus;
			err2 : OSStatus;
			found : Boolean;
			profileIndex : longint;
			thisProfileID : ICProfileID;
			thisProfileName : Str255;
			{$ifc icDebug}
				perm : ICPerm;
			{$endc}
	begin
		trueProfileID := kICNilProfileID;
		err := StartUsingIC(globals);
		if err = noErr then begin
			ICAssert( (ICGetPerm(globals^.inst, perm) = noErr) & (perm = icNoPerm) );
			
			err := ICBegin(globals^.inst, icReadOnlyPerm);
			if err = noErr then begin
			
				found := false;
				profileIndex := 1;
				repeat
					err := ICGetIndProfile(globals^.inst, profileIndex, thisProfileID);
					if err = noErr then begin
						err := ICGetProfileName(globals^.inst, thisProfileID, thisProfileName);
					end; (* if *)
					if err = noErr then begin
						found := ProfileNamesEqual(profileName, thisProfileName);
						if not found then begin
							profileIndex := profileIndex + 1;
						end; (* if *)
					end; (* if *)
				until found | (err <> noErr);
				
				if found then begin
					trueProfileID := thisProfileID;
				end; (* if *)
				
				err2 := ICEnd(globals^.inst);
				if err = noErr then begin
					err := err2;
				end; (* if *)
			end; (* if *)
			
		end; (* if *)
		ResolveSettingByName := err;
	end; (* ResolveSettingByName *)
	
	function ResolveSetting(globals : ComponentGlobalsPtr; setting : ICSettingHandle;
							var trueProfileID : ICProfileID) : OSStatus;
		var
			err : OSStatus;
			tmpName : Str255;
	begin
		trueProfileID := kICNilProfileID;
		err := StartUsingIC(globals);
		if err = noErr then begin
			err := ICGetProfileName(globals^.inst, setting^^.profileID, tmpName);
			if err = noErr then begin
				// We found it by ID, return the ID and we're done.
				trueProfileID := setting^^.profileID;
			end else if err = icProfileNotFoundErr then begin
			
				// We did not find it by ID, let's try looking for it by name.
				// But we only do this if the script codes match, ie the setting
				// was made on a system with the same system script as we're
				// currently running.  Otherwise we end up comparing
				// text returned by ICGetProfileName (which is, by definition,
				// in the system script), with text from another script system.
				
				if setting^^.profileNameScript = GetScriptManagerVariable(smSysScript) then begin
					err := ResolveSettingByName(globals,
												setting^^.profileName,
												trueProfileID);
				end; (* if *)
			end; (* if *)
		end; (* if *)
		ResolveSetting := err;
	end; (* ResolveSetting *)
	
	(* ***** Location Manager Routines ***** *)

	function ALMGetCurrent(globals : ComponentGlobalsPtr; setting : ICSettingHandle) : ComponentResult;
		var
			err : OSStatus;
			s : SInt8;
	begin
		err := StartUsingIC(globals);
		
		// First, resize the handle to accept our setting.
		
		if err = noErr then begin
			SetHandleSize(Handle(setting), sizeof(ICSetting));
			err := MemError;
		end; (* if *)
		
		// Then fill out the fields in the handle.
		
		if err = noErr then begin
			s := HGetState(Handle(setting));
			HLock(Handle(setting));
			
			// First set up default values, in case we fail later.
			
			setting^^.magic := kICLMMCreator;
			setting^^.version := kCurrentSettingVersion;
			setting^^.profileID := kICNilProfileID;
			setting^^.profileName := '';
			setting^^.profileNameScript := GetScriptManagerVariable(smSysScript);
			
			// Now get the profile ID and name from IC.
			
			err := ICGetCurrentProfile(globals^.inst, setting^^.profileID);
			if err = noErr then begin
				err := ICGetProfileName(globals^.inst, setting^^.profileID, setting^^.profileName);
			end; (* if *)

			HSetState(Handle(setting), s);
		end; (* if *)
		
		{$ifc icDebug}
			if err = noErr then begin
				ICAssert( ValidateSetting(setting) = noErr );
			end; (* if *)
		{$endc}
		
		ALMGetCurrent := err;
	end; (* ALMGetCurrent *)
	
	function ALMSetCurrent(globals : ComponentGlobalsPtr; setting : ICSettingHandle; var flags : ALMRebootFlags) : ComponentResult;
		var
			err : OSStatus;
			trueProfileID : ICProfileID;
			currentProfileID : ICProfileID;
	begin
		flags := kALMNoChange;
		
		// Validity check setting.

		err := ValidateSetting(setting);

		// Resolve setting down to profile ID, then switch in the new
		// setting.
		
		if err = noErr then begin
			err := StartUsingIC(globals);
		end; (* if *)
		if err = noErr then begin
			err := ResolveSetting(globals, setting, trueProfileID);
		end; (* if *)		
		if err = noErr then begin
			err := ICGetCurrentProfile(globals^.inst, currentProfileID);
		end; (* if *)
		if err = noErr then begin
			if trueProfileID = currentProfileID then begin
				// We're done.  We've already set flags to "no change",
				// so we just return noErr;
			end else begin
				err := ICSetCurrentProfile(globals^.inst, trueProfileID);
				if err = noErr then begin
					flags := kALMAvailableNow;
				end; (* if *)
			end; (* if *)
		end; (* if *)
		
		ALMSetCurrent := err;
	end; (* ALMSetCurrent *)
	
	function ALMCompareSetting(globals : ComponentGlobalsPtr; 
								setting1 : ICSettingHandle; 
								setting2 : ICSettingHandle; 
								var equal : Boolean) : ComponentResult;
		var
			err : OSStatus;
			name1 : Str255;
			name2 : Str255;
	begin
		{$unused globals}
		err := ValidateSetting(setting1);
		if err = noErr then begin
			err := ValidateSetting(setting1);
		end; (* if *)
		if err = noErr then begin
			name1 := setting1^^.profileName;
			name2 := setting2^^.profileName;
			equal := (setting1^^.profileID = setting2^^.profileID) &
						(setting1^^.profileNameScript = setting2^^.profileNameScript) &
						ProfileNamesEqual(name1, name2);
		end; (* if *)
		ALMCompareSetting := err;
	end; (* ALMCompareSetting *)
	
	function ALMEditSetting(globals : ComponentGlobalsPtr; setting : ICSettingHandle) : ComponentResult;
		var
			err : OSStatus;
	begin
		{$unused setting}
		
		err := StartUsingIC(globals);
		if err = noErr then begin
			err := ICEditPreferences(globals^.inst, '');
			
			// Map error codes to their more specific fake error codes,
			// so that our describe error routine knows which specific
			// case to describe.
			
			case err of
				memFullErr:
					err := memFullFakeErr;
				afpItemNotFound:
					err := afpItemNotFoundFakeErr;
				noPortErr:
					err := noPortFakeErr;
				otherwise	
					(* do nothing *) ;
			end; (* case *)
		end; (* if *)

		// Note that IC will not bring Location Manager back to front when editing
		// is done, which the documentation says it should do.  However, this is
		// too tricky for me to organise at this stage.
		
		ALMEditSetting := err;
	end; (* ALMEditSetting *)
	
	function ALMDescribeSettings(globals : ComponentGlobalsPtr; 
									setting : ICSettingHandle; 
									text : CharsHandle) : ComponentResult;
		var
			err : OSStatus;
			resourceInfo : integer;
			descriptionString : Str255;
	begin
		err := BeginUsingResources(globals, resourceInfo);
		if err = noErr then begin
			err := ValidateSetting(setting);
			if err = noErr then begin
				descriptionString := concat(
										GetIndStr(rMiscStrings, strInternetConfigSet),
										setting^^.profileName);
				SetHandleSize(Handle(text), length(descriptionString));
				err := MemError;
			end; (* if *)
			if err = noErr then begin
				BlockMoveData(@descriptionString[1], text^, length(descriptionString));
			end; (* if *)
		end; (* if *)

		// Clean up.
				
		EndUsingResources(resourceInfo);
		
		ALMDescribeSettings := err;
	end; (* ALMDescribeSettings *)
	
	function ALMDescribeError(globals : ComponentGlobalsPtr; lastErr : OSErr; var errStr : Str255) : ComponentResult;
		var
			err : OSStatus;
			stringIndex : integer;
			resourceInfo : integer;
	begin
		err := BeginUsingResources(globals, resourceInfo);
		if err = noErr then begin
			case OSStatus(lastErr) of
				badComponentInstance :
					stringIndex := strNoICErr;
				badComponentSelector :
					stringIndex := strInsufficientICErr;
				memFullFakeErr : 
					stringIndex := strNoMemoryErr;
				afpItemNotFoundFakeErr :
					stringIndex := strCantFindICErr;
				noPortFakeErr :
					stringIndex := strCantHackIt;
				icProfileNotFoundErr :
					stringIndex := strProfileNotFoundErr;
				invalidSettingFakeErr :
					stringIndex := strInvalidSetting;
				settingTooOldFakeErr :
					stringIndex := strSettingTooOld;
				settingTooNewFakeErr :
					stringIndex := strSettingTooNew;
				otherwise
					err := paramErr;		// Ask Location Manager to come up with an error.
			end; (* case *)
		end; (* if *)
		if err = noErr then begin
			errStr := GetIndStr(rMiscStrings, stringIndex);
		end; (* if *)
		
		EndUsingResources(resourceInfo);

		ALMDescribeError := err;
	end; (* ALMDescribeError *)
	
	function ALMImportExport(globals : ComponentGlobalsPtr; import : Boolean; setting : ICSettingHandle; resRefNum : integer) : ComponentResult;
		var
			err : OSStatus;
	begin
		{$unused globals}
		{$unused import}
		{$unused setting}
		{$unused resRefNum}

		// Do nothing.  The correct thing to do is to to import/export
		// the entire IC database at this point, but it's way to much work
		// for this version.  This call is optional according to the
		// Location Manager documentation, so we're just not implementing
		// it at the moment.

		err := noErr;
		ALMImportExport := err;
	end; (* ALMImportExport *)
	
	function ALMGetScriptInfo(globals : ComponentGlobalsPtr; var info : ALMScriptMgrInfo) : ComponentResult;
		var
			err : OSStatus;
	begin
		{$unused globals}
		{$unused info}

		// Do nothing.  This call is optional according to the
		// Location Manager documentation, which describes the defaults you
		// get if you don't implement it.  These defaults are fine by me,
		// so I'm just not implementing it at the moment.

		err := noErr;
		ALMGetScriptInfo := err;
	end; (* ALMGetScriptInfo *)
	
	function ALMGetInfo(globals : ComponentGlobalsPtr; var text : CharsHandle; var style : StScrpHandle; filter : ModalFilterUPP) : ComponentResult;
		var
			err : OSStatus;
			tmp : Handle;
			resourceInfo : integer;
	begin
		{$unused filter}

		err := BeginUsingResources(globals, resourceInfo);
		if err = noErr then begin
			tmp := Get1Resource('TEXT', rInfoText);
			err := CheckResError(tmp);
			if err = noErr then begin
				DetachResource(tmp);
				ICAssert(ResError = noErr);
				text := CharsHandle(tmp);
			end; (* if *)
		end; (* if *)
		if err = noErr then begin
			tmp := Get1Resource('styl', rInfoText);
			err := CheckResError(tmp);
			if err = noErr then begin
				DetachResource(tmp);
				ICAssert(ResError = noErr);
				style := StScrpHandle(tmp);
			end; (* if *)
		end; (* if *)
		
		// Clean up.
		
		EndUsingResources(resourceInfo);

		ALMGetInfo := err;
	end; (* ALMGetInfo *)

	(* ***** Component Entry Point ***** *)

	function Main(var params : ComponentParameters; storage : Ptr) : ComponentResult;
		(* The main component entry point.  It's just a standard component dispatcher. *)
		var
			oldA4 : longint;
			oldA5 : longint;
			proc: ProcPtr;
			result : ComponentResult;
	begin
		(*	Bring in the correct A4 world.  Also, set up A5 to point to the current
			application's A5 world.  We don't use A5 based globals, except insofar as
			system API routines use them.  So we don't need to create our own A5 world.  
			But we do need to make sure our A5 world is correct, because we may call 
			system routines that need A5 to point to some valid A5 world.
		*)
		oldA4 := SetCurrentA4;
		oldA5 := SetCurrentA5;

		(* Setup proc based on which routine is to be called. *)
		
		proc := nil;
		case params.what of
			// Component Manager stuff
			
			kComponentVersionSelect: 
				result := 0;			// 
			kComponentCanDoSelect: 
				proc := @ICLMMCanDo;
			kComponentOpenSelect: 
				proc := @ICLMMOpen;
			kComponentCloseSelect: 
				proc := @ICLMMClose;
			kComponentTargetSelect: 
				proc := @ICLMMTarget;
				
			// Stuff specific to Location Manager Modules

			kALMGetCurrentSelect:
				proc := @ALMGetCurrent;
			kALMSetCurrentSelect:
				proc := @ALMSetCurrent;
			kALMCompareSettingSelect:
				proc := @ALMCompareSetting;
			kALMEditSettingSelect:
				proc := @ALMEditSetting;
			kALMDescribeSettingsSelect:
				proc := @ALMDescribeSettings;
			kALMDescribeErrorSelect:
				proc := @ALMDescribeError;
			// kALMImportExportSelect:					// Need to implement this! 
			//	proc := @ALMImportExport;
			//	kALMGetScriptInfoSelect:
			//		proc := @ALMGetScriptInfo;
			kALMGetInfoSelect:
				proc := @ALMGetInfo;
			
			otherwise
				result := badComponentSelector;
		end; (* case *)

		(* If we've got a routine to call, call it.  Note that the default case
			above has set result up to return the appropriate error code if
			proc is nil.
		*)
		if proc <> nil then begin
			result := CallComponentFunctionWithStorage(Handle(storage), params, proc);
		end; (* if *)

		(* Clean up A5 and A4. *)

		oldA5 := SetA5(oldA5);
		oldA4 := SetA4(oldA4);
		Main := result;
	end; (* Main *)

end. (* ICComponent *)